@@ -1,4 +1,4 @@
-# Copyright (c) 2016-2024 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2016-2025 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -53,7 +53,8 @@ from asyncssh.public_key import get_default_x509_certificate_algs
 
 from .server import Server, ServerTestCase
 
-from .util import asynctest, patch_extra_kex, patch_getnameinfo, patch_gss
+from .util import asynctest, patch_extra_kex, patch_getaddrinfo
+from .util import patch_getnameinfo, patch_gss
 from .util import gss_available, nc_available, x509_available
 
 
@@ -281,6 +282,17 @@ class _InternalErrorClient(asyncssh.SSHClient):
         raise RuntimeError('Exception handler test')
 
 
+class _ClientCleanupError(asyncssh.SSHClient):
+    """Test of exception during client cleanup"""
+
+    def connection_lost(self, exc):
+        """Raise an error when a client is cleaned up"""
+
+        # pylint: disable=unused-argument
+
+        raise RuntimeError('Exception in cleanup test')
+
+
 class _TunnelServer(Server):
     """Allow forwarding to test server host key request tunneling"""
 
@@ -338,16 +350,16 @@ class _InvalidAuthBannerServer(Server):
         return False
 
 
-class _VersionRecordingClient(asyncssh.SSHClient):
-    """Client for testing custom client version"""
+class _AuthBannerRecordingClient(asyncssh.SSHClient):
+    """Client which records auth banner"""
 
     def __init__(self):
-        self.reported_version = None
+        self.auth_banner = None
 
     def auth_banner_received(self, msg, lang):
         """Record the client version reported in the auth banner"""
 
-        self.reported_version = msg
+        self.auth_banner = msg
 
 
 class _VersionReportingServer(Server):
@@ -1718,6 +1730,13 @@ class _TestConnection(ServerTestCase):
         with self.assertRaises(RuntimeError):
             await self.create_connection(_InternalErrorClient)
 
+    @asynctest
+    async def test_client_cleanup_error(self):
+        """Test error in client cleanup"""
+
+        async with self.connect(client_factory=_ClientCleanupError):
+            pass
+
 
 @patch_extra_kex
 class _TestConnectionNoStrictKex(ServerTestCase):
@@ -2523,11 +2542,26 @@ class _TestInvalidAuthBanner(ServerTestCase):
 
     @asynctest
     async def test_invalid_auth_banner(self):
-        """Test server sending invalid auth banner"""
+        """Test server sending invalid UTF-8 in auth banner"""
 
         with self.assertRaises(asyncssh.ProtocolError):
             await self.connect()
 
+    @asynctest
+    async def test_invalid_auth_banner_error_handler(self):
+        """Test error handler for invalid UTF-8 in auth banner"""
+
+        for errors in ('ignore', 'replace', 'backslashreplace',
+                       'surrogateescape'):
+            conn, client = \
+                await self.create_connection(_AuthBannerRecordingClient,
+                                             utf8_decode_errors=errors)
+
+            async with conn:
+                self.assertEqual(client.auth_banner,
+                                 b'\xff'.decode('utf-8', errors))
+
+
 
 class _TestExpiredServerHostCertificate(ServerTestCase):
     """Unit tests for expired server host certificate"""
@@ -2566,11 +2600,11 @@ class _TestCustomClientVersion(ServerTestCase):
         """Check custom client version"""
 
         conn, client = \
-            await self.create_connection(_VersionRecordingClient,
+            await self.create_connection(_AuthBannerRecordingClient,
                                          client_version=version)
 
         async with conn:
-            self.assertEqual(client.reported_version, 'SSH-2.0-custom')
+            self.assertEqual(client.auth_banner, 'SSH-2.0-custom')
 
     @asynctest
     async def test_custom_client_version(self):
@@ -2671,3 +2705,124 @@ class _TestListenerContextManager(ServerTestCase):
             async with asyncssh.connect('127.0.0.1', listen_port,
                                         known_hosts=(['skey.pub'], [], [])):
                 pass
+
+
+@patch_getaddrinfo
+class _TestCanonicalizeHost(ServerTestCase):
+    """Test hostname canonicalization"""
+
+    @classmethod
+    async def start_server(cls):
+        """Start an SSH server to connect to"""
+
+        return await cls.create_server(_TunnelServer)
+
+    @asynctest
+    async def test_canonicalize(self):
+        """Test hostname canonicalization"""
+
+        async with self.connect('testhost', known_hosts=None,
+                                canonicalize_hostname=True,
+                                canonical_domains=['test']) as conn:
+            self.assertEqual(conn.get_extra_info('host'), 'testhost.test')
+
+    @asynctest
+    async def test_canonicalize_max_dots(self):
+        """Test hostname canonicalization exceeding max_dots"""
+
+        async with self.connect('testhost.test', known_hosts=None,
+                                canonicalize_hostname=True,
+                                canonicalize_max_dots=0,
+                                canonical_domains=['test']) as conn:
+            self.assertEqual(conn.get_extra_info('host'), 'testhost.test')
+
+    @asynctest
+    async def test_canonicalize_ip_address(self):
+        """Test hostname canonicalization with IP address"""
+
+        async with self.connect('127.0.0.1', known_hosts=None,
+                                canonicalize_hostname=True,
+                                canonicalize_max_dots=3,
+                                canonical_domains=['test']) as conn:
+            self.assertEqual(conn.get_extra_info('host'), '127.0.0.1')
+
+    @asynctest
+    async def test_canonicalize_proxy(self):
+        """Test hostname canonicalization with proxy"""
+
+        with open('config', 'w') as f:
+            f.write('UserKnownHostsFile none\n')
+
+        async with self.connect('testhost', config='config',
+                                tunnel=f'localhost:{self._server_port}',
+                                canonicalize_hostname=True,
+                                canonical_domains=['test']) as conn:
+            self.assertEqual(conn.get_extra_info('host'), 'testhost.test')
+
+    @asynctest
+    async def test_canonicalize_always(self):
+        """Test hostname canonicalization for all connections"""
+
+        with open('config', 'w') as f:
+            f.write('UserKnownHostsFile none\n')
+
+        async with self.connect('testhost', config='config',
+                                tunnel=f'localhost:{self._server_port}',
+                                canonicalize_hostname='always',
+                                canonical_domains=['test']) as conn:
+            self.assertEqual(conn.get_extra_info('host'), 'testhost.test')
+
+    @asynctest
+    async def test_canonicalize_failure(self):
+        """Test hostname canonicalization failure"""
+
+        with self.assertRaises(socket.gaierror):
+            await self.connect('unknown', known_hosts=(['skey.pub'], [], []),
+                               canonicalize_hostname=True,
+                               canonical_domains=['test'])
+
+    @asynctest
+    async def test_canonicalize_failed_no_fallback(self):
+        """Test hostname canonicalization"""
+
+        with self.assertRaises(OSError):
+            await self.connect('unknown', known_hosts=(['skey.pub'], [], []),
+                               canonicalize_hostname=True,
+                               canonical_domains=['test'],
+                               canonicalize_fallback_local=False)
+
+    @asynctest
+    async def test_cname_returned(self):
+        """Test hostname canonicalization with cname returned"""
+
+        async with self.connect('testcname',
+                                known_hosts=(['skey.pub'], [], []),
+                                canonicalize_hostname=True,
+                                canonical_domains=['test'],
+                                canonicalize_permitted_cnames= \
+                                    [('*.test', '*.test')]) as conn:
+            self.assertEqual(conn.get_extra_info('host'), 'cname.test')
+
+    @asynctest
+    async def test_cname_not_returned(self):
+        """Test hostname canonicalization with cname not returned"""
+
+        async with self.connect('testcname',
+                                known_hosts=(['skey.pub'], [], []),
+                                canonicalize_hostname=True,
+                                canonical_domains=['test'],
+                                canonicalize_permitted_cnames= \
+                                    ['*.xxx:*.test']) as conn:
+            self.assertEqual(conn.get_extra_info('host'), 'testcname.test')
+
+    @asynctest
+    async def test_bad_cname_rules(self):
+        """Test hostname canonicalization with bad cname rules"""
+
+        with self.assertRaises(ValueError):
+            await self.connect('testcname',
+                               known_hosts=(['skey.pub'], [], []),
+                               canonicalize_hostname=True,
+                               canonical_domains=['test'],
+                               canonicalize_permitted_cnames= \
+                                   ['*.xxx:*.test:*.xxx'])
